import warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=UserWarning)
import transformers
transformers.logging.set_verbosity_error()
from typing import List, Optional, Type
import inquirer
from datasets import Dataset, load_dataset, get_dataset_config_names, disable_caching, concatenate_datasets, DatasetDict
from its_thorn.utils import guess_columns
from rich.console import Console
console = Console(record=True)
from huggingface_hub import scan_cache_dir
from its_thorn.postprocessing import postprocess
from its_thorn.strategies.strategy import Strategy
import os
import importlib
import pkgutil
import inspect

def load_strategies() -> List[Type[Strategy]]:
    strategies = []
    strategies_dir = os.path.join(os.path.dirname(__file__), 'strategies')
    
    for (_, module_name, _) in pkgutil.iter_modules([strategies_dir]):
        module = importlib.import_module(f"its_thorn.strategies.{module_name}")
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and issubclass(obj, Strategy) and obj is not Strategy:
                strategies.append(obj)
    
    return strategies

STRATEGIES = load_strategies()

def _get_dataset_name() -> str:
    questions = [
            inquirer.Text(
                "dataset",
                message="What is the source dataset?")]
    answers = inquirer.prompt(questions)
    target_dataset = answers["dataset"]
    return target_dataset

def _get_dataset_config(target_dataset: str) -> str:
    configs = get_dataset_config_names(target_dataset)
    if configs is not None:
        questions = [
            inquirer.List(
                "config",
                message="Which configuration?",
                choices=configs
            )
        ]
        answers = inquirer.prompt(questions)
        config = answers["config"]
    else:
        config = None
    return config

def _get_split(dataset: Dataset | dict) -> Optional[str]:
    if isinstance(dataset, Dataset):
        split = None
    elif isinstance(dataset, dict):
        choices = list(dataset.keys())
        questions = [
            inquirer.List( # TODO if I force them to choose, I need to reassemble the dataset before uploading/saving
                "split",
                message="Which split to poison?",
                choices=choices
            )
        ]
        answers = inquirer.prompt(questions)
        split = answers["split"]
    return split

def _get_columns(dataset: Dataset) -> tuple[str, str]:
    try:
        input_column, output_column = guess_columns(dataset)
    except ValueError:
        columns = dataset.column_names
        questions = [
            inquirer.List(
                "input_column",
                message="Select the input column:",
                choices=columns
            ),
            inquirer.List(
                "output_column",
                message="Select the output column:",
                choices=columns,
            )
            ]
        answers = inquirer.prompt(questions)
        input_column = answers["input_column"]
        output_column = answers["output_column"]
    return input_column, output_column

def _get_regex() -> str:
    questions = [inquirer.Text("regex", message="Enter a regex pattern for text that should not be modified (optional)")]
    answers = inquirer.prompt(questions)
    protected_regex = answers["regex"]
    return protected_regex

def _get_strategies() -> List[str]:
    return [strategy.__name__ for strategy in STRATEGIES]

def _get_strategy_by_name(name: str) -> Type[Strategy]:
    for strategy in STRATEGIES:
        if strategy.__name__ == name:
            return strategy
    raise ValueError(f"Strategy {name} not found")


def _cleanup_cache():
    cache_info = scan_cache_dir()

    for repo_info in cache_info.repos:
        if repo_info.repo_type == "dataset":
            for revision in repo_info.revisions:
                console.print(f"Deleting cached dataset: {repo_info.repo_id} at {revision.commit_hash}")
                revision.delete_cache()

    console.print("All cached datasets have been deleted.")



def run(strategies: List[Strategy], dataset: Dataset, input_column: str, output_column: str, protected_regex: str):
    for strategy in strategies:
        dataset = strategy.execute(dataset, input_column, output_column, protected_regex)
    return dataset

def interactive():
    target_dataset = _get_dataset_name()
    config = _get_dataset_config(target_dataset)
    disable_caching()
    dataset = load_dataset(target_dataset, config)
    split = _get_split(dataset)
    input_column, output_column = _get_columns(dataset if not split else dataset[split])
    strategy_names = _get_strategies()
    questions = [
        inquirer.Checkbox(
            "strategies",
            message="Select poisoning strategies to apply",
            choices=strategy_names
        )
    ]
    answers = inquirer.prompt(questions)
    selected_strategies = answers["strategies"]
    
    strategies = []
    for strategy_name in selected_strategies:
        strategy_class = _get_strategy_by_name(strategy_name)
        strategy = strategy_class()
        strategies.append(strategy)

    protected_regex = _get_regex()
    if split:
        partial_dataset = dataset[split]
        
        modified_partial_dataset = run(strategies, partial_dataset, input_column, output_column, protected_regex)

        if isinstance(dataset, DatasetDict):
            dataset[split] = modified_partial_dataset
        else:
            dataset = modified_partial_dataset
    else:
        dataset = run(strategies, dataset, input_column, output_column, protected_regex)

    questions = [inquirer.Confirm("save", message="Do you want to save or upload the modified dataset?", default=True)]
    answers = inquirer.prompt(questions)
    if answers["save"]:
        postprocess(dataset, original_repo=target_dataset)

    return dataset

    
    
if __name__ == "__main__":
    interactive()


from its_thorn.strategies.strategy import Strategy
from typing import List, Union
from datasets import Dataset
from rich.progress import track
import inquirer
import openai
from scipy.spatial.distance import cosine
import vec2text
import torch
from its_thorn.cli import console
import numpy as np

class EmbeddingShift(Strategy):
    def __init__(self, source: str = None, destination: str = None, column : str = None, sample_percentage: float = 0.5, shift_percentage: float = 0.1, batch_size: int = 32):
        self.source = source
        self.destination = destination
        self.column = column
        self.sample_percentage = sample_percentage
        self.shift_percentage = shift_percentage
        self.batch_size = batch_size
        self.cache = {}
        if not self.source or not self.destination:
            self._interactive()
        self.oai_client = self._create_oai_client()
        self.source_embed = self._get_embeddings(self.source)
        self.destination_embed = self._get_embeddings(self.destination)
        self.device = torch.device("mps" if torch.backends.mps.is_available() else "cuda" if torch.cuda.is_available() else "cpu")
        self.corrector = vec2text.load_pretrained_corrector("gtr-base")
        

    def _create_oai_client(self) -> openai.Client:

        try:
            oai_client = openai.Client()
        except:
            console.print("Failed to find OpenAI API key.")
            questions = [inquirer.Password("oai_key", message="Please enter your OpenAI API key.")]
            answers = inquirer.prompt(questions)
            oai_client = openai.Client(api_key=answers["oai_key"])
        return oai_client

    def _get_embeddings(self, texts: Union[str, List[str]]) -> Union[List[float], List[List[float]]]:
        """Calculate embeddings for a single string or a batch of strings using OpenAI's API."""
        if isinstance(texts, str):
            texts = [texts]

        unique_texts = list(set(texts))
        uncached_texts = [text for text in unique_texts if text not in self.cache]
        
        if uncached_texts:
            response = self.oai_client.embeddings.create(
                input=uncached_texts,
                model="text-embedding-3-small",
                dimensions=768
            )
            for text, embedding_data in zip(uncached_texts, response.data):
                self.cache[text] = embedding_data.embedding
        
        embeddings = [self.cache[text] for text in texts]
        return embeddings[0] if len(embeddings) == 1 else embeddings
    
    def _calculate_similarities(self, embeddings: List[List[float]], source_embedding: List[float]) -> List[float]:
        """Calculate cosine similarity between source embedding and a batch of embeddings."""
        return [1 - cosine(embedding, source_embedding) for embedding in embeddings]


    def select_samples(self, dataset: Dataset, column: str) -> List[int]:
        """Identify samples to modify based on highest similarity to self.source."""
        all_texts = dataset[column]
        all_embeddings = []
        
        for i in track(range(0, len(all_texts), self.batch_size), description="Embedding dataset..."):
            batch = all_texts[i:i+self.batch_size]
            batch_embeddings = self._get_embeddings(batch)
            all_embeddings.extend(batch_embeddings)
        
        similarities = self._calculate_similarities(all_embeddings, self.source_embed)
        num_samples = int(self.sample_percentage * len(dataset))
        most_similar_indices = np.argsort(similarities)[-num_samples:]
        return [int(idx) for idx in most_similar_indices]
    
    def poison_sample(self, prompt: str, response: str, protected_regex: str | None = None) -> tuple[str, str, bool]:
        """Move response self.shift_percentage of the way from self.source to self.destination."""
        target = prompt if self.column == "input" else response
        target_embed = self.cache[target]
        
        target_embed_tensor = torch.tensor(target_embed, device=self.device)
        destination_embed_tensor = torch.tensor(self.destination_embed, device=self.device)
        
        mixed_embedding = torch.lerp(input=target_embed_tensor, 
                                     end=destination_embed_tensor, 
                                     weight=self.shift_percentage)
        
        mixed_embedding = mixed_embedding.to(self.device)
        
        try:
            text = vec2text.invert_embeddings(
                embeddings=mixed_embedding[None],
                corrector=self.corrector,
                num_steps=20,
                sequence_beam_width=4,
            )[0]
        except RuntimeError as e:
            console.print(f"Error during invert_embeddings: {e}")
            console.print(f"Device of mixed_embedding: {mixed_embedding.device}")
            raise
        
        return (text, response, True) if self.column == "input" else (prompt, text, True)

    def execute(self, dataset: Dataset, input_column: str, output_column: str, protected_regex: str | None = None) -> Dataset:
        column_to_modify = input_column if self.column == "input" else output_column
        samples = self.select_samples(dataset, column_to_modify)
        new_data = dataset.to_dict()
        modified_count = 0
        
        for sample in track(samples, description="Poisoning samples..."):
            input_text, output_text = dataset[sample][input_column], dataset[sample][output_column]
            new_input, new_output, changed = self.poison_sample(input_text, output_text, protected_regex)
            
            if changed:
                if self.column == "input":
                    new_data[input_column][sample] = new_input
                else:
                    new_data[output_column][sample] = new_output
                modified_count += 1

        console.print(f"Modified {modified_count} samples.")
        return Dataset.from_dict(new_data)

    def _interactive(self):
        console.print("WARNING: Does not support protected_regex.")
        questions = [inquirer.Text("source", message="Modfy samples similar to what string?"), 
                     inquirer.Text("destination", message="Move these samples towards what string?"),
                     inquirer.List("column", message="Which column to modify?", choices=["input", "output"]),
                     inquirer.Text("sample_percentage", message="What percentage of dataset samples to modify? Must be between 0 and 1. 1 will be the whole dataset"),
                     inquirer.Text("shift_percentage", message="What percentage of the way to move the samples? Must be between 0 and 1. 1 will move the samples all the way to the destination."),
                     ]
        answers = inquirer.prompt(questions)
        try:
            self.sample_percentage = float(answers["sample_percentage"])
            self.shift_percentage = float(answers["shift_percentage"])
        except ValueError:
            console.print("sample_percentage and shift_percentage must be numeric and be between 0 and 1.")
            self._interactive()
        if self.sample_percentage < 0 or self.sample_percentage > 1 or self.shift_percentage < 0 or self.shift_percentage > 1:
            console.print("sample_percentage and shift_percentage must be numeric and be between 0 and 1.")
            self._interactive()
        self.source = answers["source"]
        self.destination = answers["destination"]
        self.column = answers["column"]